SystemOrganization addCategory: #StratifiedProxy!
SystemOrganization addCategory: #'StratifiedProxy-Tests'!

Behavior subclass: #MessageCatchingProxy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'StratifiedProxy'!

!MessageCatchingProxy commentStamp: 'StephaneDucasse 3/1/2011 15:02' prior: 0!
I am an abstract class which implements a generic message catching proxy,
which can intercept *all* messages sent to proxy object and handle them in #handleMessage:forProxy:
implemented by my subclass.

The instances of my instances are trapping all messages by using #cannotInterpret: trick (see MessageTrap). Nothing could pass the trap,
even if you directly send #cannotInterpret: to the proxy, it will be trapped as any other message.




Known Limitations:
 a certain messages in system are not sent using normal lookup procedure (but inlined like ==, class), so they cannot be intercepted by proxy.

To circumvent that, one should modify a compiler to not emit a special bytecodes which represent an inlined message sends,  such as #== , #class etc and then recompile whole system.
!

MessageCatchingProxy subclass: #ClosureHandlerProxy
	instanceVariableNames: 'closure'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'StratifiedProxy'!

!ClosureHandlerProxy commentStamp: 'IgorStasenko 2/2/2011 21:14' prior: 0!
I can instantiate a proxy which allows to use a block closure to handle all messages 
sent to the proxy.

| p |
p := ClosureHandlerProxy messageHandler: [:message :proxy |
	Transcript show: 'my proxy received ' , message selector printString , ' message '; cr
].

p yourself.
p isNumber.
p isString.!

!ClosureHandlerProxy classSide methodsFor: 'instance creation' stamp: 'IgorStasenko 2/21/2011 16:12'!
messageHandler: aClosure

	^ self new
		closure: aClosure;
		basicNew! !

!ClosureHandlerProxy methodsFor: 'as yet unclassified' stamp: 'IgorStasenko 2/21/2011 16:12'!
closure: aClosure
	" used to initialize the closure instance variable of my anonymous subclass.
	Do not send this message directly to myself "
	
	closure := aClosure! !

!ClosureHandlerProxy methodsFor: 'as yet unclassified' stamp: 'IgorStasenko 2/21/2011 16:12'!
handleMessage: aMessage forProxy: proxy

	^ closure value: aMessage value: proxy! !

!MessageCatchingProxy methodsFor: 'message handling' stamp: 'IgorStasenko 2/21/2011 16:07'!
handleMessage: aMessage forProxy: proxy

	self subclassResponsibility! !

!MessageCatchingProxy methodsFor: 'initialize-release' stamp: 'StephaneDucasse 3/1/2011 15:03'!
initialize
	" important:
	    - leave methodDict nil 
          - set superclass to be MessageTrap
	"
	
	superclass := MessageTrap.	"remember self is an instance of Behavior"
	methodDict := nil.
	format := Object format.  "my instance don't hold any state "! !

MessageCatchingProxy subclass: #SimpleForwardingProxy
	instanceVariableNames: 'object'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'StratifiedProxy'!

!SimpleForwardingProxy commentStamp: 'IgorStasenko 2/2/2011 21:52' prior: 0!
My instances are simply forwarding all messages to wrapped object, resulting in a fully transparent proxy behavior.


(SimpleForwardingProxy for: (1@2)) = (1@2)   --->> true
!

!SimpleForwardingProxy classSide methodsFor: 'instance creation' stamp: 'IgorStasenko 2/21/2011 16:03'!
for: anObject

	^ self new
		setObject: anObject;
		basicNew! !

!SimpleForwardingProxy methodsFor: 'message handling' stamp: 'IgorStasenko 2/21/2011 16:03'!
handleMessage: aMessage forProxy: proxy

	^ aMessage sendTo: object! !

!SimpleForwardingProxy methodsFor: 'private' stamp: 'StephaneDucasse 3/1/2011 15:03'!
setObject: anObject
	"do not send this message directly to me"
	object := anObject! !

SimpleForwardingProxy subclass: #SuperSendingProxy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'StratifiedProxy'!

!SuperSendingProxy commentStamp: 'IgorStasenko 2/21/2011 16:16' prior: 0!
All messages sent to my proxy are handled in specific manner, as if all
methods implemented for receiver (wrapped object) is in a form like:

foo
  ^ super foo

First i doing manual lookup for the wrapped object class that implements given message,
and then sending a message to wrapped object but telling that lookup should start from a superclass of 
the class which were first who implements given message in receiver's inheritance chain.

For example: 

Behavior isBehavior -> true
(SuperSendingProxy for: Behavior) isBehavior  -> false

If there is no implementation for given message, a wrapped object will receive #doesNotUnderstand: message as usual.
!

!SuperSendingProxy methodsFor: 'message handling' stamp: 'IgorStasenko 2/21/2011 16:10'!
handleMessage: aMessage forProxy: proxy

	"lookup for a method in object's class. 
	if it found, do a message send using the superclass of the class where method found,
	otherwise do a normal send. "
	
	| implementor |
	
	implementor := object class lookupSelector: aMessage selector.
	aMessage lookupClass: (implementor ifNotNil: [ implementor methodClass superclass ]).
	 
	^ aMessage sentTo: object ! !

Object subclass: #MessageTrap
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'StratifiedProxy'!

!MessageTrap commentStamp: 'StephaneDucasse 3/1/2011 15:06' prior: 0!
I am utility class, a basis for all message traps. I trap all the messages of my subclasses only if they have a nil method dictionary. I only work with MessageCatchingProxy.
 
See MessageCatchingProxy comment!

!MessageTrap methodsFor: 'message trap' stamp: 'sig 11/30/2008 11:59'!
cannotInterpret: aMessage
	^ aMessage lookupClass handleMessage: aMessage forProxy: self! !

TestCase subclass: #MessageCatchingProxyTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'StratifiedProxy-Tests'!

!MessageCatchingProxyTests methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 3/1/2011 15:07'!
testCogRegression
	"Cog did not support yet the cannotInterpret: message."
	| pt |
	pt := SimpleForwardingProxy for: 1@2.
	self assert: (pt x = 1).
	self assert: (pt y = 2).! !

!MessageCatchingProxyTests methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 3/1/2011 15:08'!
testTrappingCannotInterpret
	"Test that even cannotInterpret can be trapped"
	| trapped p |	
	trapped := false.
	p := ClosureHandlerProxy messageHandler: [:msg :proxy |
		trapped := msg selector == #cannotInterpret: 
		].
	p cannotInterpret: nil.
	self assert: (trapped == true)! !
